#!/usr/bin/env python3

# Copyright 2021 Canonical Limited.
#
# This file is part of juju-verify.
#
# juju-verify is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# juju-verify is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see https://www.gnu.org/licenses/.

"""Entrypoint to the 'juju-verify' plugin."""
import argparse
import asyncio
import logging
import os
import sys
import typing
from typing import Tuple, Union

from juju import errors
from juju.model import Model

from juju_verify import logger as juju_verify_logger
from juju_verify import stream_handler
from juju_verify.exceptions import CharmException, JujuVerifyError, VerificationError
from juju_verify.utils.unit import find_units, find_units_on_machine
from juju_verify.verifiers import SUPPORTED_CHARMS, BaseVerifier, get_verifiers
from juju_verify.verifiers.result import set_stop_on_failure

# set MAX_FRAME_SIZE to 64MB to connect python-libjuju to the model
JUJU_MAX_FRAME_SIZE = 2**26
logger = logging.getLogger(__name__)


async def connect_model(model_name: Union[str, None]) -> Model:
    """Connect to a custom or default Juju model.

    The Juju model can be identified by 'model_name' or the currently active
    model will be used if left unspecified.
    """
    model = Model()
    try:
        if model_name:
            logger.debug("Connecting to model '%s'.", model_name)
            await model.connect(
                model_name=model_name, max_frame_size=JUJU_MAX_FRAME_SIZE
            )
        else:
            logger.debug("Connecting to currently active model.")
            await model.connect(max_frame_size=JUJU_MAX_FRAME_SIZE)
    except errors.JujuError as exc:
        raise CharmException(
            f"Failed to connect to the model.{os.linesep}{exc}"
        ) from exc

    return model


class ExtendAction(argparse.Action):  # pylint: disable=too-few-public-methods
    """Extend action for argparse.

    NOTE (rgildein): This action should be removed after python 3.6 support ends, because
                     since Python 3.8 the "extend" is available directly in stdlib.
    """

    @typing.no_type_check
    def __call__(self, parser, namespace, values, option_string=None):
        """Extend existing items with values."""
        items = getattr(namespace, self.dest) or []
        items = [*items, *values]  # extend list of items
        setattr(namespace, self.dest, items)


def parse_charm_mapping(charm_map: str = "") -> Tuple[str, str]:
    """Parse value of --map-charm argument into a tuple (app_name, charm_name).

    Expected input format of mapping is two colon separated strings.

    :param charm_map: Colon separated pair <APP_NAME>:<CHARM_NAME>.
    :return: Input value parsed into tuple (app_name, charm_name)
    """
    if not isinstance(charm_map, str):
        raise ValueError("--map-charm arguments expects string type value.")

    try:
        app_name, charm_name = charm_map.split(":")
    except ValueError as exc:
        raise ValueError(
            "Unexpected format of --map-charm argument. For more info see --help."
        ) from exc

    return app_name, charm_name


def parse_args() -> argparse.Namespace:
    """Parse cli arguments."""
    description = (
        "Verify that it's safe to perform selected action on specified units."
        f"{os.linesep}Currently supported charms are:"
    )
    description += "".join(f"\n\t* {verifier}" for verifier in SUPPORTED_CHARMS)
    parser = argparse.ArgumentParser(
        description=description, formatter_class=argparse.RawDescriptionHelpFormatter
    )
    parser.register("action", "extend", ExtendAction)

    parser.add_argument(
        "--model", "-m", required=False, help="Connect to specific model."
    )
    parser.add_argument(
        "check",
        choices=BaseVerifier.supported_checks(),
        type=str.lower,
        help="Check to verify.",
    )
    parser.add_argument(
        "-l",
        "--log-level",
        type=str.lower,
        help="Set amount of displayed information",
        default="info",
        choices=["trace", "debug", "info"],
    )
    parser.add_argument(
        "-s",
        "--stop-on-failure",
        action="store_true",
        help="Stop running checks after a failed one.",
    )
    parser.add_argument(
        "--map-charm",
        action="append",
        default=[],
        type=parse_charm_mapping,
        help="WARNING: This option can lead to failed verifications when used "
        "incorrectly. This option allows users to explicitly specify the charm used"
        " by an application. Typical use cases involve the usage of local charms or"
        " non-official charmhub repositories. Expected value format"
        " is <APP_NAME>:<CHARM_NAME>. For list of supported charms, see description"
        " in --help",
    )

    target = parser.add_mutually_exclusive_group(required=True)
    target.add_argument(
        "--units", "-u", action="extend", nargs="+", type=str, help="Units to check."
    )
    target.add_argument(
        "--machines",
        "-M",
        action="extend",
        nargs="+",
        type=str,
        help="Check all units on the machine.",
    )
    return parser.parse_args()


def config_logger(log_level: str) -> None:
    """Configure logging options."""
    log_level = log_level.lower()
    root_logger = logging.getLogger()

    if log_level == "trace":
        # 'trace' level enables debugging in juju lib and other dependencies
        stream_handler.setFormatter(
            logging.Formatter("%(asctime)s [%(levelname)s] %(name)s: %(message)s")
        )
        root_logger.setLevel(logging.DEBUG)
        juju_verify_logger.setLevel(logging.DEBUG)
    elif log_level == "debug":
        stream_handler.setFormatter(
            logging.Formatter("%(asctime)s | %(levelname)s | %(message)s")
        )
        root_logger.setLevel(logging.INFO)
        # set DEBUG level only for juju-verify logger
        juju_verify_logger.setLevel(logging.DEBUG)
    elif log_level == "info":
        stream_handler.setFormatter(logging.Formatter("%(message)s"))
        root_logger.setLevel(logging.WARNING)
        # set INFO level only for juju-verify logger
        juju_verify_logger.setLevel(logging.INFO)
    else:
        raise JujuVerifyError(f"Unsupported log level requested: '{log_level}'")


def entrypoint() -> None:
    """Execute 'juju-verify' command."""
    try:
        args = parse_args()
        set_stop_on_failure(args.stop_on_failure)
        config_logger(args.log_level)  # update logging option
        loop = asyncio.get_event_loop()
        model = loop.run_until_complete(connect_model(args.model))

        if args.units:
            units = loop.run_until_complete(find_units(model, args.units))
        elif args.machines:
            units = loop.run_until_complete(find_units_on_machine(model, args.machines))
        else:
            raise JujuVerifyError(
                "juju-verify must target either juju units or juju machines"
            )

        for verifier in get_verifiers(units, args.map_charm):
            result = verifier.verify(args.check)
            logger.info("%s", result)
    except (
        JujuVerifyError,
        CharmException,
        VerificationError,
        NotImplementedError,
    ) as exc:
        logger.error(exc)
        sys.exit(1)



# Copyright 2021 Canonical Limited.
#
# This file is part of juju-verify.
#
# juju-verify is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# juju-verify is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see https://www.gnu.org/licenses/.
"""Helper function to manage Juju action."""
from juju.action import Action

from juju_verify.utils.cache import Cache, CacheManager


def data_from_action(action: Action, key: str, default: str = "") -> str:
    """Extract value from Action.data['results'] dictionary.

    :param action: juju.Action instance
    :param key: key to search for in action's results
    :param default: default value to return if the 'key' is not found
    :return: value from the action's results identified by 'key' or default
    """
    return action.data.get("results", {}).get(key, default)


cache_manager = CacheManager(enabled=True)
cache = Cache(128)



# Copyright 2021 Canonical Limited.
#
# This file is part of juju-verify.
#
# juju-verify is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# juju-verify is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see https://www.gnu.org/licenses/.
"""Helper function to manage Juju unit."""
import asyncio
import logging
import os
import re
from typing import Any, Dict, List, Optional, Set

from juju.action import Action
from juju.application import Application
from juju.errors import JujuError
from juju.model import Model
from juju.unit import Unit

from juju_verify.exceptions import CharmException, JujuActionFailed, VerificationError
from juju_verify.utils.action import cache, cache_manager

logger = logging.getLogger(__name__)
CHARM_URL_PATTERN = re.compile(r"^(.*):(.*/)?(?P<charm>.*)(-\d+)$")


def get_cache_key(unit: Unit, action: str, **params: Any) -> int:
    """Create hash key from unit, action and params."""
    return hash(
        hash(unit.entity_id) + hash(action) + hash(tuple(sorted(params.items())))
    )


def run_command_on_unit(unit: Unit, command: str, use_cache: bool = True) -> Action:
    """Run command on unit.

    Execute is same as `juju run --unit <unit> -- <command>`
    """
    with cache_manager(use_cache):
        key = get_cache_key(unit, command)
        if key not in cache or not cache_manager.active:
            try:
                logger.debug("run command `%s` on unit %s", command, unit.entity_id)
                loop = asyncio.get_event_loop()
                result = loop.run_until_complete(unit.run(command, timeout=2 * 60))
                cache[key] = result  # save result to cache
            except JujuError as error:
                juju_error_message = os.linesep.join(f"  {err}" for err in error.errors)
                raise CharmException(
                    f"{unit.entity_id}: command `{command}` failed with errors:"
                    f"{os.linesep}{juju_error_message}"
                ) from error

        return cache[key]


async def _run_action(
    unit: Unit,
    action: str,
    params: Optional[Dict[str, Any]] = None,
    use_cache: bool = True,
) -> Action:
    """Run Juju action and wait for results."""
    params = params or {}
    key = get_cache_key(unit, action, **params)
    with cache_manager(use_cache):
        if key not in cache or not cache_manager.active:
            try:
                logger.debug("run action %s on unit %s", action, unit.entity_id)
                _action = await unit.run_action(action, **params)
                result = await _action.wait()  # wait for result
                cache[key] = result  # save result to cache
            except JujuError as error:
                raise JujuActionFailed(error, unit, action, params) from error

        return cache[key]


def run_action_on_units(
    units: List[Unit],
    action: str,
    use_cache: bool = True,
    params: Optional[Dict[str, Any]] = None,
) -> Dict[str, Action]:
    """Run Juju action on specified units.

    :param units: List/Tuple of Unit object
    :param action: Action to run on units
    :param use_cache: Use the cache to gather the result of the action
    :param params: Additional parameters for the action
    :return: Dict in format {unit_id: action} where unit_ids are strings
             provided in 'units' and actions are their matching,
             juju.Action objects that have been executed and awaited.
    """
    loop = asyncio.get_event_loop()
    task_map = {
        unit.entity_id: loop.create_task(_run_action(unit, action, params, use_cache))
        for unit in units
    }
    tasks: asyncio.Future = asyncio.gather(*task_map.values())
    results: List[Action] = loop.run_until_complete(tasks)
    result_map = dict(zip(task_map.keys(), results))

    failed_actions_msg = []
    for unit_id, action_result in result_map.items():
        if action_result.status != "completed":
            failed_actions_msg.append(
                f"Action {action} (ID: {action_result.entity_id}) failed to complete "
                f"on unit {unit_id}. For more info see "
                f"'juju show-action-output {action_result.entity_id}'"
            )

        logger.debug(
            "Action %s (ID: %s) finished with status %s. Action results `%s`.",
            action,
            action_result.entity_id,
            action_result.status,
            action_result.data.get("results", {}),
        )

    if failed_actions_msg:
        raise VerificationError(os.linesep.join(failed_actions_msg))

    return result_map


def run_action_on_unit(
    unit: Unit,
    action: str,
    use_cache: bool = True,
    params: Optional[Dict[str, Any]] = None,
) -> Action:
    """Run juju action on single unit.

    For more info, see docstring for 'run_action_on_units'. The only
    difference is that this function returns Action object directly, not
    dict {unit_id: action}.
    """
    results = run_action_on_units([unit], action, use_cache, params)
    return results[unit.entity_id]


def parse_charm_name(charm_url: str) -> str:
    """Parse charm name from full charm url.

    Example: 'cs:focal/nova-compute-141' -> 'nova-compute'
    """
    match = CHARM_URL_PATTERN.match(charm_url)
    if match is None:
        raise CharmException(f"Failed to parse charm-url: '{charm_url}'")
    return match.group("charm")


def verify_charm_unit(charm_name: str, *units: Unit) -> None:
    """Verify that units are based on required charm."""
    for unit in units:
        if not parse_charm_name(unit.charm_url) == charm_name:
            raise CharmException(
                f"The unit {unit.entity_id} does not belong to the "
                f"charm {charm_name}."
            )


def get_first_active_unit(units: List[Unit]) -> Optional[Unit]:
    """Find first unit in active workload status."""
    for unit in units:
        if unit.workload_status == "active":
            return unit

    logger.debug("No active unit was found.")
    return None


def get_applications_names(model: Model, application: str) -> List[str]:
    """Get all names of application based on the same charm."""
    applications = []
    for app_name, app in model.applications.items():
        unit = get_first_active_unit(app.units)
        if unit and parse_charm_name(unit.charm_url) == application:
            applications.append(app_name)

    return applications


def get_related_charm_units_to_app(application: Application, charm: str) -> Set[Unit]:
    """Get all units for the same charm related to application.

    :param application: Juju application
    :param charm: charm name, e.g. ceph-osd
    """
    units: Set[Unit] = set()
    for relation in application.relations:
        if parse_charm_name(relation.provides.application.charm_url) == charm:
            units = units.union(relation.provides.application.units)

    return units


def find_unit_by_hostname(model: Model, hostname: str, charm: str) -> Unit:
    """Find unit by hostname."""
    for unit in model.units.values():
        if (
            unit.machine.hostname == hostname
            and parse_charm_name(unit.charm_url) == charm
        ):
            return unit

    raise CharmException(f"could not find unit with hostname `{hostname}`")


async def find_units(model: Model, units: List[str]) -> List[Unit]:
    """Return list of juju.Unit objects that match with names in 'units' parameter.

    This function will exit program with error message if any of units is not
    found in the juju model.

    :param model: Juju model to search units in
    :param units: List of unit names to search
    :return: List of matching juju.Unit objects
    """
    selected_units: List[Unit] = []

    for unit_name in units:
        unit = model.units.get(unit_name)
        if unit is None:
            raise CharmException(f"Unit '{unit_name}' not found in the model.")

        selected_units.append(unit)

    return selected_units


async def find_units_on_machine(model: Model, machines: List[str]) -> List[Unit]:
    """Find all principal units that run on selected machines.

    :param model: Juju model to search units in
    :param machines: names of Juju machines on which to search units
    :return: List of juju.Unit objects that match units running on the machines
    """
    return [
        unit
        for _, unit in model.units.items()
        if unit.machine.entity_id in machines and not unit.data.get("subordinate")
    ]



# Copyright 2021 Canonical Limited.
#
# This file is part of juju-verify.
#
# juju-verify is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# juju-verify is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see https://www.gnu.org/licenses/.

"""Package with classes implementing verification methods for various charms."""
import logging
import os
from collections import defaultdict
from typing import Iterator, List, Optional, Tuple

from juju.unit import Unit

from juju_verify.exceptions import CharmException
from juju_verify.utils.unit import parse_charm_name
from juju_verify.verifiers.base import BaseVerifier
from juju_verify.verifiers.ceph import CephMon, CephOsd
from juju_verify.verifiers.neutron_gateway import NeutronGateway
from juju_verify.verifiers.nova_compute import NovaCompute
from juju_verify.verifiers.ovn_central import OvnCentral
from juju_verify.verifiers.result import Result, Severity

logger = logging.getLogger(__name__)


SUPPORTED_CHARMS = {
    "nova-compute": NovaCompute,
    "ceph-osd": CephOsd,
    "ceph-mon": CephMon,
    "neutron-gateway": NeutronGateway,
    "ovn-central": OvnCentral,
}


def get_verifiers(
    units: List[Unit], charm_map: Optional[List[Tuple[str, str]]] = None
) -> Iterator[BaseVerifier]:
    """Implement Factory function "verifier" creator for the supplied units.

    :param units: Juju unit(s) for which you want to produce verifier
    :param charm_map: Explicit mapping of application units to specific verifier
    :return: Correct verifier for given unit(s)
    :raises CharmException: Raised if units do not belong to the same charm or
        if the charm is unsupported for verification
    """
    if not units:
        raise CharmException("List of units can not be empty when creating verifier")

    charm_map = charm_map or []
    charms = defaultdict(list)
    mapped_applications = dict(charm_map)

    for unit in units:
        unit_application = unit.data.get("application")
        if unit_application in mapped_applications:
            # Assign charm type based on explicit mapping
            charm_type = mapped_applications[unit_application]
            logger.debug(
                "Using explicitly defined charm for unit %s: %s",
                unit.entity_id,
                charm_type,
            )
        else:
            # Infer charm type based on charm URL
            charm_type = parse_charm_name(unit.data.get("charm-url", ""))
            logger.debug("Inferred charm for unit %s: %s", unit.entity_id, charm_type)
        charms[charm_type].append(unit)

    for charm, charm_units in charms.items():
        logger.info("===[%s]===", ", ".join([unit.entity_id for unit in charm_units]))
        verifier = SUPPORTED_CHARMS.get(charm)
        if verifier is None:
            supported_charms = os.linesep.join(SUPPORTED_CHARMS.keys())
            logger.error(
                "Charm '%s' is not supported by juju-verify. Supported charms:%s%s",
                charm,
                os.linesep,
                supported_charms,
            )
            continue

        exclude_affected_units = list(set(units).difference(charm_units))
        logger.debug("Initiating verifier instance of class: %s", verifier.__name__)
        yield verifier(units=charm_units, exclude_affected_units=exclude_affected_units)



# Copyright 2022 Canonical Limited.
#
# This file is part of juju-verify.
#
# juju-verify is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# juju-verify is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see https://www.gnu.org/licenses/.
"""ovn-central verification."""
import asyncio
import logging
import os
from collections import defaultdict
from typing import Any, Dict, List, NamedTuple, Optional

import yaml
from juju.unit import Unit

from juju_verify.exceptions import JujuVerifyError
from juju_verify.utils.action import data_from_action
from juju_verify.utils.unit import run_action_on_units
from juju_verify.verifiers.base import BaseVerifier
from juju_verify.verifiers.result import Result, Severity, checks_executor

logger = logging.getLogger(__name__)

StatusCollectionType = Dict[str, "UnitClusterStatus"]


class ClusterStatus:  # pylint: disable=R0902
    """Object representation of OVN cluster status.

    Attributes of this object represent status of a single (Southbound or Northbound)
    cluster as retrieved from 'cluster-status' action of ovn-central charm unit. Example
    of an input data for initialization of this class:

    .. code-block:: yaml

        Cluster ID: 567e7225-369e-40d6-abf8-9b442bbcd18b
        Server ID: 16335def-c21e-404c-b123-8337b3013c07
        Address: ssl:10.5.2.232:6644
        Status: cluster member
        Role: follower
        Term: 34
        Leader: dbdb
        Vote: dbdb
        Log: '[66, 66]'
        Entries not yet committed: 0
        Entries not yet applied: 0
        Servers:
        - - 7f23
          - ssl:10.75.224.151:6644
        - - 1bc5
          - ssl:10.75.224.192:6644
        - - 1ef7
          - ssl:10.75.224.120:6644
        unit_map:
          ovn-central/0: 7f23
          ovn-central/1: 1bc5
          ovn-central/2: 1ef7
    """

    def __init__(self, raw_status: str) -> None:
        """Initialize Cluster status class.

        For more info about the purpose and input data format see class documentation.

        :param raw_status: YAML string containing status of a single OVN cluster.
        """
        try:
            status_dict = yaml.safe_load(raw_status)
            if not isinstance(status_dict, dict):
                raise ValueError("Input data is not a YAML dictionary.")
        except (yaml.YAMLError, ValueError) as exc:
            raise JujuVerifyError(
                f"Failed to load OVN status as YAML:{os.linesep}{raw_status}"
            ) from exc

        try:
            self.cluster_id = status_dict["cluster_id"]
            self.server_id = status_dict["server_id"]
            self.address = status_dict["address"]
            self.status = status_dict["status"]
            self.role = status_dict["role"]
            self.term = status_dict["term"]
            self.leader = status_dict["leader"]
            self.vote = status_dict["vote"]
            self.log = status_dict["log"]
            self.entries_not_yet_committed = status_dict["entries_not_yet_committed"]
            self.entries_not_yet_applied = status_dict["entries_not_yet_applied"]
            self.servers = status_dict["servers"]
            self.unit_map = status_dict["unit_map"]

        except KeyError as exc:
            raise JujuVerifyError(
                f"Failed to deserialize OVN cluster status. Missing "
                f"'{exc.args[0]}' from cluster status data."
            ) from exc

    @property
    def short_id(self) -> str:
        """Return first 4 characters of a Server ID.

        This format is often used to refer to servers within cluster instead of full ID.
        """
        return self.server_id[:4]

    @property
    def is_leader(self) -> bool:
        """Return true if Server that reported this status consider itself a leader."""
        return self.leader == "self"

    def __eq__(self, other: Any) -> bool:
        """Implement equality comparison for ClusterStatus instances."""
        if not isinstance(other, ClusterStatus):
            return NotImplemented

        comparable_attrs = [
            "cluster_id",
            "server_id",
            "address",
            "status",
            "role",
            "term",
            "leader",
            "vote",
            "log",
            "entries_not_yet_committed",
            "entries_not_yet_applied",
            "servers",
            "unit_map",
        ]
        equal = True
        try:
            for attr in comparable_attrs:
                if self.__getattribute__(attr) != other.__getattribute__(attr):
                    equal = False
                    break
        except AttributeError:
            equal = False

        return equal


class UnitClusterStatus(NamedTuple):
    """Convenience class that groups Southbound and Northbound cluster status.

    This is how a single ovn-central unit views status of these clusters.
    """

    southbound: ClusterStatus
    northbound: ClusterStatus

    def __eq__(self, other: Any) -> bool:
        """Implement equality comparison between UnitClusterStatus instances."""
        if not isinstance(other, UnitClusterStatus):
            return NotImplemented

        return (
            self.southbound == other.southbound and self.northbound == other.northbound
        )


class OvnCentral(BaseVerifier):
    """Implementation of verification checks for the ovn-central charm."""

    NAME = "ovn-central"

    def __init__(
        self, units: List[Unit], exclude_affected_units: Optional[List[Unit]] = None
    ):
        """Initialize ovn-central charm verifier."""
        super().__init__(units, exclude_affected_units)
        self._complete_cluster_status: StatusCollectionType = {}
        self._all_application_units: List[Unit] = []

    @property
    def complete_cluster_status(self) -> StatusCollectionType:
        """Collect and return status of a cluster from every ovn-central unit.

        This method collects Southbound and Northbound cluster status reports from every
        unit that belongs to an application of currently verified unit(s).
        """
        report_err = (
            "{} failed to report {} cluster status. Please try to run "
            "`cluster-status` action manually."
        )
        if not self._complete_cluster_status:
            action_results = run_action_on_units(
                self.all_application_units, "cluster-status", use_cache=False
            )
            for unit_name, result in action_results.items():
                sb_data = data_from_action(result, "ovnsb")
                nb_data = data_from_action(result, "ovnnb")
                if not sb_data:
                    raise JujuVerifyError(report_err.format(unit_name, "Southbound"))
                if not nb_data:
                    raise JujuVerifyError(report_err.format(unit_name, "Northbound"))

                self._complete_cluster_status[unit_name] = UnitClusterStatus(
                    southbound=ClusterStatus(sb_data), northbound=ClusterStatus(nb_data)
                )

        return self._complete_cluster_status

    @property
    def all_application_units(self) -> List[Unit]:
        """Return list of all units from an application of currently verified unit(s)."""
        if not self._all_application_units:
            app = self.units[0].application
            self._all_application_units = [
                unit for unit in self.model.units.values() if unit.application == app
            ]
        return self._all_application_units

    @staticmethod
    def cluster_tolerance(size: int) -> int:
        """Return number of failed nodes that the cluster can tolerate.

        This calculation is based on the raft protocol's requirement to have "(N/2) + 1"
        active nodes to maintain quorum.

        :param size: Current size of a cluster.
        :return: Number of nodes that can fail without loosing cluster quorum
        """
        if size < 1:
            return 0

        min_quorum_size = (size // 2) + 1
        return size - min_quorum_size

    def check_single_application(self) -> Result:
        """Verify that all units that are being verified belong to the same app."""
        apps = {unit.application for unit in self.units}
        if len(apps) > 1:
            app_list = ", ".join(apps)
            return Result(
                Severity.FAIL,
                f"Can't verify multiple ovn-central application at the same time. "
                f"Currently selected units belong to: {app_list}",
            )

        return Result(Severity.OK, "Selected units are part of only one application.")

    def check_leader_consistency(self) -> Result:
        """Verify consistency of ovn-central cluster."""
        result = Result()
        leader_mappings: Dict[str, Dict[str, List[str]]] = {
            "Southbound": defaultdict(list),
            "Northbound": defaultdict(list),
        }
        for unit, cluster_data in self.complete_cluster_status.items():
            clusters = [
                ("Southbound", cluster_data.southbound),
                ("Northbound", cluster_data.northbound),
            ]
            for cluster_name, status in clusters:
                # Collect opinions on who is the leader from each unit
                leader = status.short_id if status.is_leader else status.leader
                if leader:
                    leader_mappings[cluster_name][leader].append(unit)

        # Verify that there's consensus on who is the cluster leader
        for cluster, leaders in leader_mappings.items():
            if len(leaders) > 1:
                leader_err = f"There's no consensus on {cluster} cluster leader. "
                for leader, units in leaders.items():
                    unit_list = ", ".join(units)
                    leader_err += f"{leader} is supported by {unit_list}; "
                result.add_partial_result(Severity.FAIL, leader_err)
            elif len(leaders) < 1:
                result.add_partial_result(
                    Severity.FAIL,
                    f"No unit reported elected leader in {cluster} cluster.",
                )
            else:
                leader = list(leaders.keys())[0]
                result.add_partial_result(
                    Severity.OK, f"All units agree that {leader} is {cluster} leader."
                )

        return result

    def check_uncommitted_logs(self) -> Result:
        """Verify that there are no uncommitted log entries on cluster leaders."""
        result = Result()
        for unit, cluster_data in self.complete_cluster_status.items():
            clusters = [
                ("Southbound", cluster_data.southbound),
                ("Northbound", cluster_data.northbound),
            ]
            for cluster_name, status in clusters:
                # Verify that cluster leader does not have any uncommitted entries
                if status.is_leader:
                    commits_ok = not bool(status.entries_not_yet_committed)
                    result_severity = Severity.OK if commits_ok else Severity.FAIL
                    result.add_partial_result(
                        result_severity,
                        f"{unit} ({cluster_name} leader) reports "
                        f"{status.entries_not_yet_committed} uncommitted log entries.",
                    )

        return result

    def check_unknown_servers(self) -> Result:
        """Verify that there are no servers in cluster without associated unit.

        This situation can happen if, for example, a unit is removed from the
        application, but it's unable to gracefully leave cluster.
        """
        # For this check, we don't need to go through ClusterStatus of every unit.
        # We can just randomly pick one.
        result = Result()
        err_msg = "{} cluster reports servers that are not associated with a unit."
        cluster_status = self.complete_cluster_status[self.unit_ids[0]]
        if "UNKNOWN" in cluster_status.southbound.unit_map:
            result.add_partial_result(Severity.FAIL, err_msg.format("Southbound"))

        if "UNKNOWN" in cluster_status.northbound.unit_map:
            result.add_partial_result(Severity.FAIL, err_msg.format("Northbound"))

        return result or Result(
            Severity.OK, "No disassociated cluster members reported."
        )

    def check_supported_charm_version(self) -> Result:
        """Verify that targeted application has required actions.

        This verifier requires action "cluster-status" to be present on the charm.
        """
        loop = asyncio.get_event_loop()
        app_name = self.units[0].application
        app = self.model.applications[app_name]
        all_actions = loop.run_until_complete(app.get_actions())
        if "cluster-status" in all_actions.keys():
            return Result(Severity.OK, "Charm supports all required actions.")

        return Result(
            Severity.FAIL,
            "Charm does not support required action 'cluster-status'. Please try "
            "upgrading charm.",
        )

    def check_reboot(self) -> Result:
        """Check that it's safe to temporarily bring down selected units.

        This check verifies that rebooting selected units won't bring the cluster below
        its fault tolerance.
        """
        result = Result()
        all_units = len(self.all_application_units)
        fault_tolerance = self.cluster_tolerance(all_units)
        units_to_reboot = len(self.units)

        if units_to_reboot > fault_tolerance:
            result.add_partial_result(
                Severity.FAIL,
                f"OVN cluster with {all_units} units can not tolerate simultaneous "
                f"reboot of {units_to_reboot} units.",
            )
        else:
            result.add_partial_result(
                Severity.OK,
                f"OVN cluster with {all_units} units can safely tolerate simultaneous "
                f"reboot of {units_to_reboot} units.",
            )
            if units_to_reboot == fault_tolerance:
                result.add_partial_result(
                    Severity.WARN,
                    "While the rebooted units are down, this cluster won't be able to "
                    "tolerate any more failures.",
                )

        return result

    def check_downscale(self) -> Result:
        """Check that removing selected units won't affect cluster's fault tolerance."""
        all_units = len(self.all_application_units)
        units_to_remove = len(self.units)
        original_tolerance = self.cluster_tolerance(all_units)
        post_removal_tolerance = self.cluster_tolerance(all_units - units_to_remove)
        base_msg = f"Removing {units_to_remove} units from cluster of {all_units} "

        if original_tolerance < 1:
            return Result(
                Severity.FAIL,
                f"Cluster of {all_units} units has already 0 fault tolerance.",
            )

        if post_removal_tolerance < 1:
            return Result(
                Severity.FAIL, base_msg + "would bring its fault tolerance to 0."
            )

        if original_tolerance != post_removal_tolerance:
            return Result(
                Severity.WARN,
                base_msg
                + f"will decrease its fault tolerance from {original_tolerance}"
                f" to {post_removal_tolerance}.",
            )

        return Result(Severity.OK, base_msg + "won't impact its fault tolerance.")

    def preflight_checks(self) -> Result:
        """Run common checks that verify integrity of the OVN cluster.

        These checks should be prerequisite before any further checks are run for both
        reboot and shutdown actions
        """
        charm_supported = checks_executor(self.check_supported_charm_version)
        if not charm_supported.success:
            return charm_supported

        return charm_supported + checks_executor(
            self.check_single_application,
            self.check_leader_consistency,
            self.check_uncommitted_logs,
            self.check_unknown_servers,
        )

    def verify_reboot(self) -> Result:
        """Verify that it's safe to reboot selected ovn-central units."""
        preflight_results = self.preflight_checks()
        if not preflight_results.success:
            return preflight_results

        return preflight_results + checks_executor(self.check_reboot)

    def verify_shutdown(self) -> Result:
        """Verify that it's safe to shut down selected ovn-central units."""
        preflight_results = self.preflight_checks()
        if not preflight_results.success:
            return preflight_results

        return preflight_results + checks_executor(self.check_downscale)



# Copyright 2021 Canonical Limited.
#
# This file is part of juju-verify.
#
# juju-verify is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# juju-verify is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see https://www.gnu.org/licenses/.
"""neutron-gateway verification."""

from typing import Any, Dict, List

import yaml
from juju.unit import Unit
from packaging.version import Version

from juju_verify.utils.action import data_from_action
from juju_verify.utils.unit import parse_charm_name, run_action_on_unit
from juju_verify.verifiers.base import BaseVerifier, Result, Severity
from juju_verify.verifiers.result import checks_executor


class NeutronGateway(BaseVerifier):
    """Implementation of verification checks for the neutron-gateway charm."""

    NAME = "neutron-gateway"
    action_name_result_map = {
        "show-routers": "router-list",
        "show-dhcp-networks": "dhcp-networks",
        "show-loadbalancers": "load-balancers",
    }
    action_name_failure_string_map = {
        "show-routers": "The following routers are non-redundant: {}",
        "show-dhcp-networks": "The following DHCP networks are non-redundant: {}",
        "show-loadbalancers": (
            "The following LBaasV2 LBs were found: {}. LBaasV2 does not offer HA."
        ),
    }

    @classmethod
    def get_unit_resource_list(
        cls, unit: Unit, get_resource_action_name: str
    ) -> Dict[str, Dict[str, Any]]:
        """Given a get resource action, return the relevant resources on the unit."""
        get_resource_action = run_action_on_unit(unit, get_resource_action_name)
        action_name_res = cls.action_name_result_map[get_resource_action_name]
        resource_dict_json = data_from_action(get_resource_action, action_name_res)
        resource_dict = yaml.safe_load(resource_dict_json)
        return resource_dict

    def get_all_ngw_units(self) -> List[Unit]:
        """Get all neutron-gateway units, including those not being shutdown."""
        return [
            unit
            for unit in self.model.units.values()
            if parse_charm_name(unit.data.get("charm-url", "")) == self.NAME
        ]

    def get_resource_list(self, get_resource_action_name: str) -> List[dict]:
        """Given a get resource action, return matching resources from all units."""
        resource_list = []
        shutdown_hostname_list = [unit.machine.hostname for unit in self.units]

        for unit in self.get_all_ngw_units():
            hostname = unit.machine.hostname
            host_resource_dict = self.get_unit_resource_list(
                unit, get_resource_action_name
            )

            # add host metadata to resource
            for resource_id, info in host_resource_dict.items():
                resource_list.append(
                    {
                        "id": resource_id,
                        "host": hostname,
                        "juju-entity-id": unit.entity_id,
                        "shutdown": hostname in shutdown_hostname_list,
                        **info,
                    }
                )

        return resource_list

    def get_shutdown_resource_list(self, get_resource_action_name: str) -> List[dict]:
        """Return a list of resources matching action that are going to be shutdown."""
        res_list = self.get_resource_list(get_resource_action_name)
        return [
            resource
            for resource in res_list
            if resource["shutdown"] and resource["status"] == "ACTIVE"
        ]

    def get_online_resource_list(self, get_resource_action_name: str) -> List[dict]:
        """Return a list of resources matching action, that will remain online."""
        res_list = self.get_resource_list(get_resource_action_name)
        return [
            resource
            for resource in res_list
            if not resource["shutdown"] and resource["status"] == "ACTIVE"
        ]

    def check_non_redundant_resource(self, action_name: str) -> Result:
        """Check that there are no non-redundant resources matching the resource type."""
        shutdown_resource_list = self.get_shutdown_resource_list(action_name)
        redundant_resource_list = self.get_online_resource_list(action_name)

        shutdown_resource_set = set(res["id"] for res in shutdown_resource_list)
        redundant_resource_set = set(res["id"] for res in redundant_resource_list)
        non_redundant_list = shutdown_resource_set - redundant_resource_set
        if non_redundant_list:
            failure_string = self.action_name_failure_string_map[action_name]
            reason = failure_string.format(", ".join(non_redundant_list))
            result = Result(Severity.FAIL, reason)
        else:
            resource = self.action_name_result_map[action_name]
            reason = f"Redundancy check passed for: {resource}"
            result = Result(Severity.OK, reason)
        return result

    def warn_router_ha(self) -> Result:
        """Warn that HA routers should be manually failed over."""
        result = Result()
        shutdown_resource_list = self.get_shutdown_resource_list("show-routers")

        router_failover_err_list = []
        for router in shutdown_resource_list:
            if router["ha"]:
                _id = router["id"]
                entity_id = router["juju-entity-id"]
                host = router["host"]
                error_string = f"{_id} (on {entity_id}, hostname: {host})"
                router_failover_err_list.append(error_string)

        if router_failover_err_list:
            error_string = (
                "It's recommended that you manually failover the following "
                "routers: {}"
            )
            reason = error_string.format(", ".join(router_failover_err_list))
            result = Result(Severity.WARN, reason)

        return result

    def warn_lbaas_present(self) -> Result:
        """Warn that LBaasV2 loadbalancers are present on the verified units."""
        result = Result()
        shutdown_lbaas_list = self.get_resource_list("show-loadbalancers")
        units_with_lbaas = {unit["juju-entity-id"] for unit in shutdown_lbaas_list}
        affected_lbaas_units = set(self.unit_ids) & units_with_lbaas

        if affected_lbaas_units:
            message = (
                "Following units have neutron LBaasV2 load-balancers that will be "
                "lost on unit reboot/shutdown: {}"
            )
            reason = message.format(", ".join(affected_lbaas_units))
            result = Result(Severity.WARN, reason)

        return result

    def version_check(self) -> Result:
        """Check minimum required version of Juju agents.

        NeutronGateway verifier requires that all the neutron-gateway units run juju
        agent >=2.8.10 due to reliance on juju.Machine.hostname feature.
        """
        return self.check_minimum_version(Version("2.8.10"), self.get_all_ngw_units())

    def verify_reboot(self) -> Result:
        """Verify that it's safe to reboot selected neutron-gateway units."""
        return self.verify_shutdown()

    def verify_shutdown(self) -> Result:
        """Verify that it's safe to shutdown selected neutron-gateway units."""
        version_check = checks_executor(self.version_check)
        if not version_check.success:
            return version_check

        return version_check + checks_executor(
            self.warn_router_ha,
            self.warn_lbaas_present,
            (self.check_non_redundant_resource, {"action_name": "show-routers"}),
            (self.check_non_redundant_resource, {"action_name": "show-dhcp-networks"}),
        )



# Copyright 2021 Canonical Limited.
#
# This file is part of juju-verify.
#
# juju-verify is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# juju-verify is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see https://www.gnu.org/licenses/.
"""nova-compute verification."""
import json
import logging

from juju_verify.utils.action import data_from_action
from juju_verify.utils.unit import run_action_on_unit
from juju_verify.verifiers.base import BaseVerifier
from juju_verify.verifiers.result import Result, Severity, checks_executor

logger = logging.getLogger(__name__)


class NovaCompute(BaseVerifier):
    """Implementation of verification checks for the nova-compute charm."""

    NAME = "nova-compute"

    def check_no_running_vms(self) -> Result:
        """Check that none of the units have VMs running on them."""
        result = Result()
        instance_count_action = "instance-count"
        instance_count_results = self.run_action_on_all(instance_count_action)

        for unit_id, action in instance_count_results.items():
            running_vms = int(data_from_action(action, "instance-count"))
            result.add_partial_result(
                Severity.FAIL if running_vms != 0 else Severity.OK,
                f"Unit {unit_id} is running {running_vms} VMs.",
            )

        return result

    def check_no_empty_az(self) -> Result:
        """Check that removing units wont cause empty availability zone."""

        def is_active(node: dict) -> bool:
            return node["state"] == "up" and node["status"] == "enabled"

        node_name_actions = self.run_action_on_all("node-name")
        target_nodes = [
            data_from_action(action, "node-name")
            for _, action in node_name_actions.items()
        ]

        action = run_action_on_unit(self.units[0], "list-compute-nodes")
        compute_nodes = json.loads(data_from_action(action, "compute-nodes"))

        affected_zones = {
            node["zone"] for node in compute_nodes if node["host"] in target_nodes
        }
        zones_after_change = {
            node["zone"]
            for node in compute_nodes
            if node["host"] not in target_nodes and is_active(node)
        }

        empty_zones = affected_zones - zones_after_change

        if empty_zones:
            result = Result(
                Severity.FAIL,
                f"Removing these units would leave following "
                f"availability zones empty: {empty_zones}",
            )
        else:
            result = Result(Severity.OK, "Empty Availability Zone check passed.")
        return result

    def verify_reboot(self) -> Result:
        """Verify that it's safe to reboot selected nova-compute units."""
        return checks_executor(self.check_no_running_vms, self.check_no_empty_az)

    def verify_shutdown(self) -> Result:
        """Verify that it's safe to shutdown selected nova-compute units."""
        return self.verify_reboot()



# Copyright 2021 Canonical Limited.
#
# This file is part of juju-verify.
#
# juju-verify is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# juju-verify is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see https://www.gnu.org/licenses/.
"""Helper function to manage cache."""
from collections import OrderedDict
from contextlib import _GeneratorContextManager, contextmanager
from typing import Any, Generator, List


class Cache:
    """Cache for storing outputs using specific keys.

    Usage example:
    cache = Cache()

    def run_action(name: str) -> Any:
        if name in cache:
            return cache[name]

        return _run_action(name)

    run_action('test')  # action 'test' is executed
    run_action('test')  # action 'test' is not executed
    run_action('test')  # action 'test' is not executed
    """

    def __init__(self, maxsize: int):
        """Initialize cache object."""
        self._cache: OrderedDict = OrderedDict()
        self.maxsize: int = maxsize

    def __getitem__(self, key: int) -> Any:
        """Get cached value."""
        if key in self._cache:
            self._cache.move_to_end(key)  # reorder cache

        return self._cache[key]

    def __setitem__(self, key: int, value: Any) -> None:
        """Cache the value using the key."""
        self._cache[key] = value

        # remove the oldest key
        if len(self._cache) > self.maxsize:
            oldest_key = next(iter(self._cache))
            del self._cache[oldest_key]

    def __iter__(self) -> Generator:
        """Iterate over cache keys."""
        for key in self._cache:
            yield key

    def clear(self) -> None:
        """Clear cached data."""
        self._cache.clear()

    @property
    def keys(self) -> List[Any]:
        """Return cached keys."""
        return list(self._cache.keys())


class CacheManager:
    """A cache manager that determines when to use this cache.

    Usage example:
    cache_manager = CacheManager(enabled=True)
    cache = Cache()

    def run_action(name: str) -> Any:
        if cache_manager.active and name in cache:
            return cache[name]

        return _run_action(name)

    run_action('test')  # action 'test' is executed
    run_action('test')  # action 'test' is not executed

    with cache_manager(use_cache=False):
        run_action('test')  # action 'test' is executed

    run_action('test')  # action 'test' is not executed
    """

    def __init__(self, enabled: bool = True):
        """Init the Cache class with default state."""
        self._enabled: bool = enabled
        self._active: bool = False

    def __call__(self, use_cache: bool) -> _GeneratorContextManager:
        """Return cache contextmanager if instance is called."""
        return self.cache_contextmanager(use_cache)

    @property
    def active(self) -> bool:
        """Return True if cache is activated."""
        return self._active

    @property
    def enabled(self) -> bool:
        """Return True if cache is enabled."""
        return self._enabled

    def enable(self) -> None:
        """Enable the cache."""
        self._enabled = True

    def disable(self) -> None:
        """Disable the cache."""
        self._enabled = False

    @contextmanager
    def cache_contextmanager(self, use_cache: bool) -> Generator:
        """Possibility to temporarily run the cache."""
        self._active = use_cache is True and self.enabled
        try:
            yield
        finally:
            self._active = False



# Copyright 2021 Canonical Limited.
#
# This file is part of juju-verify.
#
# juju-verify is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation, either version 3 of the License, or (at your option) any later
# version.
#
# juju-verify is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program. If not, see https://www.gnu.org/licenses/.
"""Manage package and distribution."""
from setuptools import setup
import subprocess
from typing import List


def find_version() -> str:
    """Parse juju-verify version based on the git tag."""
    try:
        cmd: List[str] = ["git", "describe", "--tags", "--always", "HEAD"]
        gitversion: str = subprocess.check_output(
            cmd, stderr=subprocess.DEVNULL
        ).decode().strip()
        if all(char.isdigit() or char == "." for char in gitversion):
            # gitversion in tagged commits comprises only of numbers and dots
            return gitversion
        else:
            # gitversion in commits that are not tagged has number of commits
            # since the last tag and commit id attached to it.
            # <tagname>-<ncommits-ahead>-<commit-id> (e.g. 0.2-8-adfebee)
            build: List[str] = gitversion.split("-")
            return "{}.post{}".format(build[0], build[1])
    except IndexError:
        cmd: List[str] = ["git", "rev-list", "--count", "HEAD"]
        commits_count: str = subprocess.check_output(
            cmd, stderr=subprocess.DEVNULL
        ).decode().strip()
        return "0.0.dev{}".format(commits_count)
    except subprocess.CalledProcessError:
        return "0.0.dev0"


setup(version=find_version())
